// kp1 --- 1bit CPU
// version 0.1
// Copyright 2005 by Yasuo Kuwahara

module kp1(clk, reset, inst_adr, inst, intreq, intack,
	data_adr, data_in, data_out, data_rd, data_we);
	input clk, reset, intreq;
	input [15:0] inst;
	input [7:0] data_in;
	output intack;
	output [14:0] inst_adr;
	output [15:0] data_adr;
	output [7:0] data_out;
	output data_rd, data_we;
	reg [14:0] pc = 0;
	reg [5:0] count = 0;
	reg s0 = 0, s1 = 0;
	reg [3:0] instl = 0;
	reg [10:0] regadr = 0;
	wire reg_a_data_in, reg_a_data_out, reg_a_we, reg_b_we, y_alu, co;
	wire normalcnt, cond;
	wire [7:0] reg_b_data_in, reg_b_data_out;
	wire fetch = ~s0 & ~s1;
	wire [5:0] insts = fetch ? inst[15:10] : { instl, regadr[10:9] };
	wire i_nop = insts[5:2] == 4'b0000;
	wire i_bank = insts[5:2] == 4'b0010;
	wire i_j = insts == 6'b000110;
	wire i_call = insts == 6'b000111;
	wire i_jr = insts[5:1] == 5'b00010;
	wire i_ldst = insts[5:3] == 3'b010 | insts[5:2] == 4'b0110;
	wire i_li = insts[5:2] == 4'b0011;
	wire i_mov = insts[5:2] == 4'b1000;
	wire i_op = insts[5];
	wire i_s = insts[5:2] == 4'b0111;
	wire i_jcond = i_j & cond;
	wire i_jncond = i_j & ~cond;
	wire i_li1 = i_li & s0;
	wire i_jr12 = i_jr & s0;
	wire i_jr3 = i_jr & s1 & ~s0;
	wire i_call2 = i_call & s1 & s0;
	wire i_call3 = i_call & s1 & ~s0;
	wire i_jrcall = i_jr | i_call;
	wire i_jrcall2 = i_jrcall & s1 & s0;
	wire i_jrcall3 = i_jrcall & s1 & ~s0;
	wire i_ldst1 = i_ldst & ~s1 & s0;
	wire i_ldst2 = i_ldst & s1 & s0;
	wire i_ldst12 = i_ldst & s0;
	wire i_ldstex = i_ldst & s1 & ~s0;
	wire i_s1 = i_s & s0;
	wire i_s2 = i_s & s1;
	wire i_op1 = i_op & ~s1 & s0;
	wire i_op2 = i_op & s1 & s0;
	wire i_op3 = i_op & s1 & ~s0;
	//
	// REGWAIT
	//
	reg regwait = 0;
	always @(posedge clk) begin
		if (regwait) regwait <= 0;
		else if (i_ldst | i_s | i_jr | normalcnt |
			i_jcond & s0 | 
			i_li1 & count == 0 |
			i_ldstex & count != 0)
			regwait <= 1;
	end
	//
	// REGADR
	//
	wire [4:0] regadr_or = { 5 { intreq } };
	always @(posedge clk) if (~regwait) begin
		if (fetch) 
			regadr <= { inst[11:7] | regadr_or, 6'b000000 };
		else if (i_s2 | i_ldst2) 
			regadr <= { inst[6:2], 6'b000000 };
		else if (i_op3) 
			regadr <= regadr + 1;
		else if (i_s1 | i_ldst1 | i_ldstex) 
			regadr <= regadr + 8;
	end
	//
	// REGADR2
	//
	reg [10:0] regadr2 = 0;
	always @(posedge clk) if (~regwait) begin
		if (fetch) 
			regadr2 <= { inst[6:2] | regadr_or, 6'b000000 };
		else if (i_s) 
			if (s1) 
				regadr2[10:8] <= reg_b_data_out[2:0];
			else
				regadr2[7:0] <= reg_b_data_out;
		else if (i_op3) 
			regadr2 <= regadr2 + 1;
		else if (i_li | i_jrcall)
			regadr2 <= regadr2 + 8;
	end
	//
	// REGBANK
	//
	reg [3:0] regbank = 0;
	always @(posedge clk) if (~regwait) begin
		if (fetch & i_bank) 
			regbank <= inst[3:0];
		if (i_jr3 & regadr[9]) 
			regbank <= reg_b_data_out[3:0];
	end
	//
	// SF
	//
	reg sf = 0;
	always @(posedge clk) if (~regwait) begin
		if (fetch) 
			sf <= 0;
		else if (i_s) 
			sf <= 1;
	end
	//
	// S0
	//
	wire gotoif = (i_ldstex | i_li1 | i_op3) & count == 0 | i_j & s0 | i_jrcall3;
	always @(posedge clk) if (~regwait) begin
		if (gotoif | i_ldst2 | i_op2 | i_op1 & i_mov | i_jrcall2)
			s0 <= 0;
		else if (fetch & (intreq | ~i_nop & ~i_bank) | i_op3)
			s0 <= 1;
	end
	//
	// S1
	//
	always @(posedge clk) if (~regwait) begin
		if (gotoif | i_s2 | i_op3) 
			s1 <= 0;
		else if (s0) 
			s1 <= 1;
	end
	//
	// GATE
	//
	function [5:0] ord;
		input [1:0] size;
		case (size)
			0: ord = 6'b111000;
			1: ord = 6'b110000;
			2: ord = 6'b100000;
			3: ord = 6'b000000;
		endcase
	endfunction
	wire s_exit = i_op3 & sf & (regadr2[5:0] | ord(inst[1:0])) == 6'b111111;
	reg gate = 0;
	always @(posedge clk) if (~regwait) begin
		if (fetch) 
			gate <= 0;
		else if (i_s2) 
			gate <= ~reg_b_data_out[7];
		else if (s_exit) 
			gate <= ~gate;
	end
	//
	// DATA0
	//
	reg data0 = 0;
	always @(posedge clk) if (~regwait) begin
		if (fetch)
			data0 <= 0;
		else if (i_op2) 
			data0 <= reg_a_data_out;
	end
	//
	// COUNT
	//
	function [5:0] bytecountdecode;
		input [1:0] a;
		case (a)
			0: bytecountdecode = 6'b000000;
			1: bytecountdecode = 6'b000001;
			2: bytecountdecode = 6'b000011;
			3: bytecountdecode = 6'b000111;
		endcase
	endfunction
	function [5:0] bitcountdecode;
		input [1:0] a;
		case (a)
			0: bitcountdecode = 6'b000111;
			1: bitcountdecode = 6'b001111;
			2: bitcountdecode = 6'b011111;
			3: bitcountdecode = 6'b111111;
		endcase
	endfunction
	always @(posedge clk) if (~regwait) begin
		if (fetch) begin
			count <= i_ldst | i_li ? bytecountdecode(inst[1:0]) : bitcountdecode(inst[1:0]);
		end
		else if (i_ldstex | i_li1 | i_op3) 
			count <= count - 1;
	end
	//
	// FILLER
	//
	reg filler = 0;
	always @(posedge clk) if (~regwait) begin
		if (fetch) 
			filler <= ~inst[7];
		else if (i_s2) 
			filler <= reg_b_data_out[6];
		else if (s_exit & filler) 
			filler <= reg_a_data_out;
	end
	//
	// FLAG
	//
	reg cy = 0, z = 0;
	wire flag_init = fetch & i_op & ~i_mov;
	wire flag_update = i_op3 & ~i_mov;
	wire flag_restore = i_jr3 & regadr[9];
	always @(posedge clk) if (~regwait) begin
		if (flag_init) 
			cy <= 0;
		else if (flag_update) 
			cy <= co;
		if (flag_restore) 
			cy <= reg_b_data_out[4];
	end
	always @(posedge clk) if (~regwait) begin
		if (flag_init) 
			z <= 1;
		else if (flag_update & y_alu) 
			z <= 0;
		if (flag_restore) 
			z <= reg_b_data_out[5];
	end
	//
	// INTACK
	//
	reg intack = 0;
	always @(posedge clk) 
		if (reset) intack <= 0;
		else if (~regwait) begin
			if (fetch & intreq) 
				intack <= 1;
			if (i_call3) 
				intack <= 0;
		end
	//
	// SYS
	//
	reg sys = 0;
	always @(posedge clk) 
		if (reset) sys <= 0;
		else if (~regwait) begin
			if (i_call3 & intack) 
				sys <= 1;
			if (i_jr3 & regadr[9]) 
				sys <= reg_b_data_out[6];
		end
	//
	// ADR
	//
	reg [15:0] adr = 0;
	always @(posedge clk) if (~regwait) begin
		if (i_ldst12) 
			if (s1) 
				adr[15:8] <= reg_b_data_out;
			else
				adr[7:0] <= reg_b_data_out;
		else if (i_ldstex) 
			adr <= adr + 1;
	end
	//
	// INSTL
	//
	always @(posedge clk) if (~regwait) begin
		if (fetch) instl <= intreq ? 4'b0001 : inst[15:12];
		else if (i_s2) instl <= 4'b1000;
	end
	//
	// PC
	//
	function condsel;
		input [2:0] i;
		input z, cy;
		case (i)
			4: condsel = ~z;
			5: condsel = z;
			6: condsel = ~cy;
			7: condsel = cy;
			default: condsel = 1;
		endcase
	endfunction
	assign cond = condsel(regadr2[9:7], z, cy);
	assign normalcnt =
		fetch & (i_nop | i_li | i_j | i_call | i_bank) & ~intreq | 
		(i_ldstex | i_op3) & count == 0 |
		i_jncond & s0 |
		i_jr3 & regadr[9] & regadr[8];
	always @(posedge clk) 
		if (reset) pc <= 0;
		else if (~regwait) begin
			if (normalcnt | i_li1 & ~regadr2[3]) 
				pc <= pc + 1;
			else if (i_jcond & s0 | i_call2) 
				pc <= intack ? 2 : inst[14:0];
			else if (i_jr12) 
				if (s1) 
					pc[14:8] <= reg_b_data_out[6:0];
				else 
					pc[7:0] <= reg_b_data_out;
		end
	//
	// REG
	//
	function [7:0] reg_b_datasel;
		input [15:0] inst;
		input ldc;
		input [14:0] pc;
		input [7:0] data;
		input bytesel;
		input filler;
		if (i_li | ldc) 
			reg_b_datasel = bytesel ? inst[15:8] : inst[7:0];
		else if (i_call)
			if (s0) reg_b_datasel = s1 ? { 1'b0, pc[14:8] } : pc[7:0];
			else reg_b_datasel = { 1'b0, sys, z, cy, regbank };
		else reg_b_datasel = data;
	endfunction
	wire [10:0] reg_ab_adr = 
		i_op & (sf ^ s1) | 
		i_li | 
		i_call | 
		i_jr ? regadr2 : regadr;
	wire [3:0] reg_u_adr = reg_ab_adr[10] ? { 3'b111, intack | sys } : regbank;
	assign reg_a_data_in = gate ? filler : y_alu;
	assign reg_a_we = i_op3 & (~instl[3] | ~instl[1] | ~instl[0]);
	assign reg_b_data_in = reg_b_datasel(inst, i_ldstex & instl[0], pc, data_in, i_ldst ? adr[0] : regadr2[3], filler);
	assign reg_b_we = ~regwait & 
		(i_ldstex & ~instl[1] | 
		i_li1 |
		i_call & (s0 | s1));
	regfile regfile(.clk(clk), 
		.reg_a_adr({ reg_u_adr, reg_ab_adr[9:0] }), 
		.reg_a_data_in(reg_a_data_in), .reg_a_data_out(reg_a_data_out), 
		.reg_a_we(reg_a_we),
		.reg_b_adr({ reg_u_adr, reg_ab_adr[9:3] }), 
		.reg_b_data_in(reg_b_data_in), .reg_b_data_out(reg_b_data_out), 
		.reg_b_we(reg_b_we));
	//
	// DATA
	//
	assign data_adr = adr;
	assign data_out = reg_b_data_out;
	assign data_rd = i_ldstex & ~instl[1] & ~instl[0];
	assign data_we = i_ldstex & instl[1] & ~regwait;
	assign inst_adr = i_ldstex & instl[0] ? adr[15:1] : pc;
	//
	// ALU
	//
	alu alu(.i(instl), .a(data0), .b(reg_a_data_out), .ci(cy), 
			.y(y_alu), .co(co));
	initial $monitor($stime,, "%x%x %x %x %x %x %x %x %x %x %x", 
		s1, s0, inst_adr, inst, 
		reg_ab_adr, reg_b_data_in, reg_b_we,
		adr, data_out, data_rd, data_we);
endmodule

module alu(i, a, b, ci, y, co);
	input [3:0] i;
	input a, b, ci;
	output y, co;
	wire sub = ~i[2] & i[1];
	wire a1 = sub ^ a;
	wire ci1 = (sub | i[2] & ~i[1] & ~i[0]) ^ ci;
	wire y1 = a1 ^ b ^ ci1;
	wire co1 = a1 & b | b & ci1 | ci1 & a1;
	assign y = i != 4'b1000 ? i[2] & (i[1] | ~i[0]) ? co1 : y1 : b;
	assign co = ~i[2] & (i[1] | i[0]) & (sub ^ co1);
endmodule

module regfile(clk, 
	reg_a_adr, reg_a_data_in, reg_a_data_out, reg_a_we,
	reg_b_adr, reg_b_data_in, reg_b_data_out, reg_b_we);
	input clk;
	input [13:0] reg_a_adr;
	input [10:0] reg_b_adr;
	input reg_a_data_in;
	input [7:0] reg_b_data_in;
	input reg_a_we, reg_b_we;
	output reg_a_data_out;
	output [7:0] reg_b_data_out;
	RAMB16_S1_S9 RAMB16_S1_S9(
		.CLKA(clk), .WEA(reg_a_we), .SSRA(1'b0), .ADDRA(reg_a_adr), 
		.DIA(reg_a_data_in), .DOA(reg_a_data_out), 
		.CLKB(clk), .WEB(reg_b_we), .SSRB(1'b0), .ADDRB(reg_b_adr), 
		.DIB(reg_b_data_in), .DIPB(1'b0), .DOB(reg_b_data_out), .DOPB());
endmodule

