module kp1_top(clk, reset, ss, led, sled, 
	ramce_n, rambe_n, ramoe_n, ramwe_n, ramadr, ramdata);
	input clk, reset;
	input [7:0] ss;
	output [7:0] led;
	output [11:0] sled;
	output [1:0] ramce_n;
	output [1:0] rambe_n;
	output ramoe_n, ramwe_n;
	output [17:0] ramadr;
	inout [15:0] ramdata;
	function [3:0] dec2;
		input [1:0] a;
		case (a)
			0: dec2 = 4'b1110;
			1: dec2 = 4'b1101;
			2: dec2 = 4'b1011;
			3: dec2 = 4'b0111;
		endcase
	endfunction
	function [7:0] sel4n;
		input [1:0] sel;
		input [7:0] a0, a1, a2, a3;
		case (sel)
			0: sel4n = ~a0;
			1: sel4n = ~a1;
			2: sel4n = ~a2;
			3: sel4n = ~a3;
		endcase
	endfunction
	wire [9:0] iadr;
	wire [15:0] idata;
	wire [15:0] data_adr;
	wire [7:0] data_in, data_out;
	wire data_rd, data_wr;
	//
	reg [7:0] outport[0:7];
	always @(posedge clk) 
		if (reset) begin
			outport[0] <= 0;
			outport[1] <= 0;
			outport[2] <= 0;
			outport[3] <= 0;
			outport[4] <= 0;
			outport[5] <= 0;
			outport[6] <= 0;
			outport[7] <= 0;
		end
		else if (data_wr & data_adr[15]) 
			outport[data_adr[2:0]] <= data_out;
	//
	reg [15:0] count;
	always @(posedge clk) count <= count + 1;
	assign sled = { dec2(count[15:14]), sel4n(count[15:14], outport[0], outport[1], outport[2], outport[3]) };
	assign led = outport[4];
	//
	reg intreq = 0;
	wire intack;
	always @(posedge clk) begin
		if (count == 16'hffff) intreq <= 1;
		else if (intack) intreq <= 0;
	end
	kp1 kp1(.clk(clk), .reset(reset), .inst_adr(iadr), .inst(idata), 
		.intreq(intreq), .intack(intack),
		.data_adr(data_adr), 
		.data_in(data_in), .data_out(data_out), 
		.data_rd(data_rd), .data_we(data_wr));
	RAMB16_S18 rom(.CLK(clk), .WE(1'b0), .SSR(1'b0), .ADDR(iadr), 
		.DI(16'h0000), .DIP(2'h0), .DO(idata));
	//
	// RAM
	//
	assign ramce_n = { 1'b1, data_adr[15] };
	assign rambe_n = 2'b10;
	assign ramoe_n = ~data_rd;
	assign ramwe_n = ~data_wr;
	assign ramadr = { 3'b000, data_adr[14:0] };
	assign ramdata[15:8] = 8'hZZ;
	assign ramdata[7:0] = data_wr ? data_out : 8'hZZ;
	assign data_in = data_adr[15] ? ss : ramdata[7:0];
endmodule

